"""Region hierarchy tree."""

import copy
import json
import logging
import re

import numpy as np
import pandas as pd

from voxcell.exceptions import VoxcellError

L = logging.getLogger(__name__)


class Matcher:
    """Helper class for value search."""
    def __init__(self, value, ignore_case=False):
        """Init Matcher."""
        self.value = value
        if isinstance(value, str):
            self.ignore_case = ignore_case
            if value.startswith("@"):
                self.value = re.compile(value[1:], re.IGNORECASE if ignore_case else 0)
        else:
            if ignore_case:
                L.warning("Not a string value; ignoring 'ignore_case' flag")

    def __call__(self, value):
        """Return True if the given value matches."""
        if hasattr(self.value, 'match'):
            return bool(self.value.search(value))

        if isinstance(value, str) and self.ignore_case:
            return self.value.upper() == value.upper()

        return self.value == value


class RegionMap:
    """Region ID <-> attribute mapping."""
    def __init__(self):
        """Init RegionMap."""
        self._data = {}
        self._children = {}
        self._parent = {}

    def get(self, _id, attr, with_ascendants=False):
        """Get attribute value associated with region ID.

        Args:
            _id (int): region ID of interest
            attr (str): attribute of interest
            with_ascendants (bool): collect attribute value upwards the "lineage"

        Returns:
            - if `with_ascendants=False`: attribute value for given region ID
            - otherwise: list of values starting from the "bottom" hierarchy level towards "top"

        Raises:
            - VoxcellError if either region ID or attribute key are can not be found

        Example:
            >>> rmap.get(315, 'name')
            'Isocortex'
        """
        if with_ascendants:
            return [self._get(k, attr) for k in self._ascendants(_id)]

        return self._get(_id, attr)

    def find(self, value, attr, ignore_case=False, with_descendants=False):
        """Find IDs of the regions matching a given attribute.

        Args:
            value: attribute value to match
            attr (str): attribute of interest
            ignore_case (bool): ignore case (when comparing strings)
            with_descendants (bool): collect region IDs downwards the "lineage"

        If `value` starts with '@' symbol, `value[1:]` is used a regular expression.
        Any substring matching the regular expression would be matched;
        please used '^' and '$' for "starts with" or "ends with" restrictions.

        Regular expressions can be used together with `ignore_case`.

        Returns:
            - if `with_descendants=False`: set of IDs of the regions matching the attribute
            - otherwise: set of region IDs matching the attribute + all their children recursively

        Example:
            >>> rmap.find("@layer 1", attr='name', ignore_case=True, with_descendants=True)
            set([1, 2, 4, 5])
        """
        matcher = Matcher(value, ignore_case=ignore_case)

        result = set()
        for _id in self._data:
            if matcher(self._get(_id, attr)):
                if with_descendants:
                    result.update(self._descendants(_id))
                else:
                    result.add(_id)

        return result

    def is_leaf_id(self, _id):
        """Indicate whether or not the input identifier is a leaf of the hierarchy tree.

        A leaf identifier is the identifier of a region with no children.

        Args:
            _id(int): region identifier, i.e., an 'id' value in hierarchy.json.

        Returns:
            True, if is a leaf, False otherwise.

        Raises:
            VoxcellError if the identifier cannot be found.

        Example:
            >>> rmap.is_leaf_id(399)
            True
            >>> rmap.is_leaf_id(-10)
            VoxcellError: Region ID not found: -10
        """
        if _id not in self._data:
            raise VoxcellError(f"Region ID not found: {_id}")
        return not self._children[_id]

    def as_dataframe(self):
        """Converts a region_map to a dataframe.

        Returns:
            pd.DataFrame with an index of the id of the node,
            and columns based on the data within the map, and a parent_id

        Note: the 'root' node should have a parent value of -1
        """
        ret = pd.DataFrame.from_dict(self._data, orient='index').set_index('id')
        parents = {k: v if v is not None else -1 for k, v in self._parent.items()}
        ret.loc[:, 'parent_id'] = pd.DataFrame.from_dict(parents, orient='index')
        ret.loc[:, 'children_count'] = [len(self._children[_id]) for _id in ret.index.to_list()]
        return ret

    @classmethod
    def from_dataframe(cls, hierarchy_df):
        """Converts a DataFrame to a region_map.

        Note: the 'root' node should have a parent value of -1.

        Note: if it is possible to cast all non-null values of a column with float dtype to int,
            then it will be done.
        """
        return cls.from_dict(_dataframe_to_dict(hierarchy_df))

    def _get(self, _id, attr):
        """Fetch attribute value for a given region ID."""
        if _id not in self._data:
            raise VoxcellError(f"Region ID not found: {_id}")
        node = self._data[_id]
        if attr not in node:
            raise VoxcellError(f"Attribute not found: '{attr}' [region ID = {_id}]")
        return node[attr]

    def _ascendants(self, _id):
        """List of ascendants for a given region ID (itself included; sorted "upwards")."""
        x = _id
        result = []
        while x is not None:
            result.append(x)
            x = self._parent[x]
        return result

    def _descendants(self, _id):
        """Set of descendants for a given region ID (itself included)."""
        result = set([_id])
        for c in self._children[_id]:
            result.update(self._descendants(c))
        return result

    @classmethod
    def from_dict(cls, d):
        """Construct RegionMap from a hierarchical dictionary."""
        def include(data, parent_id):
            # pylint: disable=protected-access,missing-docstring
            _id = data['id']
            if _id in result._data:
                raise VoxcellError(f"Duplicate id: {_id}")
            children = data.pop('children', [])
            result._data[_id] = data
            result._parent[_id] = parent_id
            result._children[_id] = [c['id'] for c in children]
            for c in children:
                include(c, _id)
        result = cls()
        include(copy.deepcopy(d), None)
        return result

    def as_dict(self):
        """Converts a region_map to a dict."""
        root_idx = None
        for k, v in self._parent.items():
            if v is None:
                root_idx = k
                break

        def create_node(key):
            return copy.deepcopy(self._data[key])

        def add_children(data, key):
            data["children"] = []
            for i in self._children[key]:
                new_node = create_node(i)
                add_children(new_node, i)
                data["children"].append(new_node)

        res = create_node(root_idx)
        add_children(res, root_idx)
        return res

    @classmethod
    def load_json(cls, filepath):
        """Construct RegionMap from JSON file.

        Note:
            If top-most object contains 'msg' field, Allen Brain Institute JSON layout is assumed.
        """
        with open(filepath, 'r', encoding='utf-8') as f:
            content = json.load(f)

        if 'msg' in content:
            if len(content['msg']) > 1:
                raise VoxcellError("Unexpected JSON layout (more than one 'msg' child)")
            content = content['msg'][0]

        return cls.from_dict(content)


def _dataframe_to_dict(hierarchy_df):
    """Use a dataframe to create a dict that can then be used by RegionMap.from_dict()."""
    nodes = hierarchy_df.to_dict(orient="index")
    float_cols = hierarchy_df.dtypes.loc[hierarchy_df.dtypes == float].index.to_list()
    dropna_float_cols = {
        float_col: hierarchy_df[float_col].dropna()
        for float_col in float_cols
    }
    float_int_cols = {
        float_col
        for float_col, col in dropna_float_cols.items()
        if (col.astype(int) == col).all()
    }
    root_idx = None
    for k, v in nodes.items():
        v["id"] = k
        v.pop("children_count", None)
        parent_id = v.pop("parent_id", None)
        for float_col in float_cols:
            if float_col in v:
                if np.isnan(v[float_col]):
                    v[float_col] = None
                elif float_col in float_int_cols:
                    v[float_col] = int(v[float_col])
        if parent_id == -1:
            if root_idx is not None:
                msg = (
                    f"Only one node can be the root node with parent_id == -1 but the node "
                    f"{root_idx} was already defined as root"
                )
                raise RuntimeError(msg)
            root_idx = k
            if "children" not in v:
                v["children"] = []
            continue
        parent_node = nodes[parent_id]
        if "children" not in parent_node:
            parent_node["children"] = []
        parent_node["children"].append(v)

    # Here the root element is extracted since each element is referenced at both the root of
    # the dict and in the children of another element
    return nodes[root_idx]



from pathlib import Path
from unittest.mock import patch

import numpy as np
import numpy.testing as npt
import pytest

import voxcell.nexus.voxelbrain as test_module
from voxcell import OrientationField, RegionMap, VoxelData
from voxcell.exceptions import VoxcellError

DATA_PATH = Path(__file__).parent / 'data'
DUMMY_ATLAS = test_module.Atlas.open('/foo')


def test_open_atlas():
    assert isinstance(test_module.Atlas.open('foo'), test_module.LocalAtlas)
    assert isinstance(test_module.Atlas.open('file://foo'), test_module.LocalAtlas)
    with pytest.raises(VoxcellError):
        test_module.Atlas.open('foo://bar')
    with pytest.raises(VoxcellError):
        test_module.Atlas.open('http://foo/bar')


@patch('os.path.exists', return_value=True)
@patch('voxcell.RegionMap.load_json', return_value='test')
def test_load_region_map(mock, _):
    actual = DUMMY_ATLAS.load_region_map()
    assert actual == 'test'
    mock.assert_called_with('/foo/hierarchy.json')


@patch('os.path.exists', return_value=True)
@patch('voxcell.nexus.voxelbrain.open')
def test_load_metadata(open_mock, _):
    with open(str(Path(DATA_PATH, 'metadata.json')), 'r') as metadata_file:
        open_mock.return_value = metadata_file
        metadata = DUMMY_ATLAS.load_metadata()
        if 'layers' in metadata:
            assert all(key in metadata['layers'] for key in ['names', 'queries', 'attribute'])


@patch('os.path.exists', return_value=True)
@patch('voxcell.VoxelData.load_nrrd', return_value='test')
def test_load_data_1(mock, _):
    actual = DUMMY_ATLAS.load_data('bar')
    assert actual == 'test'
    mock.assert_called_with('/foo/bar.nrrd')


@patch('os.path.exists', return_value=True)
@patch('voxcell.OrientationField.load_nrrd', return_value='test')
def test_load_data_2(mock, _):
    actual = DUMMY_ATLAS.load_data('bar', cls=OrientationField)
    assert actual == 'test'
    mock.assert_called_with('/foo/bar.nrrd')


@patch('os.path.exists', return_value=True)
@patch('voxcell.VoxelData.load_nrrd')
@patch('voxcell.RegionMap.load_json')
def test_get_region_mask(mock_rmap, mock_data, _):
    mock_rmap.return_value = RegionMap.from_dict({
        'id': 1,
        'acronym': 'A',
        'name': 'aA',
        'children': [{
            'id': 2,
            'acronym': 'B',
            'name': 'Bb',
        }]
    })
    mock_data.return_value = VoxelData(np.array([1, 0, 2]), [1.], [0.])

    npt.assert_equal(
        DUMMY_ATLAS.get_region_mask('A').raw,
        [True, False, True]
    )
    npt.assert_equal(
        DUMMY_ATLAS.get_region_mask('a', ignore_case=True).raw,
        [True, False, True]
    )
    with pytest.raises(VoxcellError):
        DUMMY_ATLAS.get_region_mask('a')
    npt.assert_equal(
        DUMMY_ATLAS.get_region_mask('A', with_descendants=False).raw,
        [True, False, False]
    )
    npt.assert_equal(
        DUMMY_ATLAS.get_region_mask('aa', attr='name',
                                   ignore_case=True).raw,
        [True, False, True]
    )
    with pytest.raises(VoxcellError):
        DUMMY_ATLAS.get_region_mask('aa')
    npt.assert_equal(
        DUMMY_ATLAS.get_region_mask('aA', attr='name').raw,
        [True, False, True]
    )
    with pytest.raises(VoxcellError):
        DUMMY_ATLAS.get_region_mask('aa')


expected_data = (('layer 1', {1140, 1125}),
                 ('layer 2', {1141, 1126}),
                 ('layer 3', {517, 1142, 1127}),
                 ('layer 4', {1128}),
                 ('layer 5', {1129}),
                 ('layer 6', {1130}),
                 ('L1', {1125}),)
expected_names = [v[0] for v in expected_data]
expected_ids = [v[1] for v in expected_data]


def test_get_layer():
    atlas = test_module.Atlas.open(str(DATA_PATH))
    for i in range(7):
        layer, ids = atlas.get_layer(i)
        assert expected_names[i] == layer
        assert expected_ids[i] == ids


@patch('voxcell.nexus.voxelbrain.LocalAtlas.load_region_map', return_value={})
@patch('voxcell.nexus.voxelbrain.LocalAtlas.load_metadata')
def test_get_layer_ids_raise(load_metadata_mock, _):
    atlas = test_module.Atlas.open(str(DATA_PATH))
    load_metadata_mock.return_value = {'layers': {'name': ['layer1', 'layer4']}}
    with pytest.raises(VoxcellError):
        atlas.get_layer(0)

    load_metadata_mock.return_value = \
        {'layers': {'names': ['layer1', 'layer4'], 'queries': ['@.*1$', '@.*4$']}}
    with pytest.raises(VoxcellError):
        atlas.get_layer(1)

    load_metadata_mock.return_value = \
        {'layers': {'names': ['layer1', 'layer4'], 'queries': ['@.*4$'], 'attribute': 'name'}}
    with pytest.raises(VoxcellError):
        atlas.get_layer(2)


def test_get_layer_ids():
    atlas = test_module.Atlas.open(str(DATA_PATH))
    names, ids = atlas.get_layers()
    npt.assert_equal(names, expected_names)
    for id_, expected in zip(ids, expected_ids):
        assert id_ == expected


@patch('voxcell.nexus.voxelbrain.LocalAtlas.load_data')
def test_get_layer_volume(load_data_mock):
    atlas = test_module.Atlas.open(str(DATA_PATH))
    load_data_mock.return_value = VoxelData(
        np.array([[[1125], [1126], [1130]]]), voxel_dimensions=(1.0, 1.0, 1.0))

    layer_volume = atlas.get_layer_volume()
    # In metadata.json, 'L1' and '@.*1' defines the same set of ids.
    # L1 wins because it comes at the end.
    npt.assert_array_equal(layer_volume.raw, np.array([[[7], [2], [6]]]))



import numpy as np
from numpy.testing import assert_array_equal

from voxcell import vector_fields as test_module


def test_combine_vector_fields_empty():
    assert_array_equal(test_module.combine_vector_fields([]), np.empty((0,)))


def test_combine_vector_fields_low_dim_2():
    r = test_module.combine_vector_fields([np.ones((2, 2)) * 1,
                                           np.ones((2, 2)) * 2])

    assert r.shape == (2, 2, 2)


def test_combine_vector_fields_simple_0():
    # no rotation
    right = [1, 0, 0]
    up = [0, 1, 0]
    fwd = [0, 0, 1]

    r = test_module.combine_vector_fields([
            np.array(right * np.prod((2, 2, 2))).reshape((2, 2, 2, 3)),
            np.array(up * np.prod((2, 2, 2))).reshape((2, 2, 2, 3)),
            np.array(fwd * np.prod((2, 2, 2))).reshape((2, 2, 2, 3))
        ])

    assert r.shape == (2, 2, 2, 3, 3)

    assert_array_equal(r[1, 1, 1], np.array([[1, 0, 0],
                                             [0, 1, 0],
                                             [0, 0, 1]]))

    assert_array_equal(np.dot(r[0, 0, 0], np.array([1, 0, 0])), np.array([1, 0, 0]))


def test_combine_vector_fields_simple_1():
    # 90 degree rotation around Y
    right = [0, 0, -1]
    up = [0, 1, 0]
    fwd = [1, 0, 0]

    r = test_module.combine_vector_fields([
            np.array(right * np.prod((2, 2, 2))).reshape((2, 2, 2, 3)),
            np.array(up * np.prod((2, 2, 2))).reshape((2, 2, 2, 3)),
            np.array(fwd * np.prod((2, 2, 2))).reshape((2, 2, 2, 3))
        ])

    assert r.shape == (2, 2, 2, 3, 3)

    assert_array_equal(r[1, 1, 1], np.array([[0, 0, 1],
                                             [0, 1, 0],
                                             [-1, 0, 0]]))

    assert_array_equal(np.dot(r[1, 1, 1], np.array([1, 0, 0])), np.array([0, 0, -1]))


def test_combine_vector_fields_high_dim():
    r = test_module.combine_vector_fields([np.ones((2, 2, 2, 2, 4)) * 1,
                                           np.ones((2, 2, 2, 2, 4)) * 2,
                                           np.ones((2, 2, 2, 2, 4)) * 3,
                                           np.ones((2, 2, 2, 2, 4)) * 4])

    assert r.shape == (2, 2, 2, 2, 4, 4)


def test_join_vector_fields_1_empty():
    assert (test_module.join_vector_fields(np.array([[0]])) ==
            np.array([[0]]))


def test_join_vector_fields_1_full():
    assert (test_module.join_vector_fields(np.array([[1]])) ==
            np.array([[1]]))


def test_join_vector_fields_2():
    assert (test_module.join_vector_fields(np.array([[1]]), np.array([[0]])) ==
            np.array([[1]]))

    assert (test_module.join_vector_fields(np.array([[0]]), np.array([[1]])) ==
            np.array([[1]]))


def test_join_vector_fields_2_override():
    assert (test_module.join_vector_fields(np.array([[1]]), np.array([[2]])) ==
            np.array([[2]]))


def test_gaussian_filter_0():
    assert_array_equal(test_module.gaussian_filter(np.array([[1, 0, 0], [1, 0, 0]]), 1),
                       np.array([[1, 0, 0], [1, 0, 0]]))


def test_gaussian_filter_1():
    assert_array_equal(test_module.gaussian_filter(np.array([[10, 0, 0], [0, 10, 0]]), 5),
                       np.array([[5, 4, 0], [4, 5, 0]]))



"""Helper mathematical functions."""

import functools
import math

import numpy as np
from scipy.spatial.transform import Rotation


def gcd(a, b):
    """Return greatest common divisor."""
    return math.gcd(a, b)


def lcm(a, b):
    """Return lowest common multiple."""
    return a * b // gcd(a, b)


# TODO consider making this a np ufunc
def lcmm(args):
    """Return lcm of args."""
    return functools.reduce(lcm, args)


def minimum_aabb(mask):
    """Calculate the minimum axis-aligned bounding box for a volume mask.

    Returns:
        A tuple containing the minimum x,y,z and maximum x,y,z
    """
    idx = np.nonzero(mask)
    return np.min(idx, axis=1), np.max(idx, axis=1)


def positions_minimum_aabb(positions):
    """Calculate the minimum axis-aligned bounding box for a list of positions.

    Returns:
        A tuple containing the minimum x,y,z and maximum x,y,z
    """
    return np.min(positions, axis=0), np.max(positions, axis=0)


def clip(mask, aabb):
    """Take a numpy array and clip it to an axis-aligned bounding box.

    Args:
        mask: numpy array
        aabb: tuple of two sets of coordinates indicating, respectively,
            the lowest and highest values for each dimension

    Returns:
        A new numpy array containing the same values as mask for the space defined by aabb
    """
    idx = tuple(slice(s, e + 1) for s, e in zip(*aabb))
    return mask[idx].copy()


def is_diagonal(matrix):
    """Check if the matrix is diagonal."""
    return np.all(matrix == np.diag(matrix.diagonal()))


def angles_to_matrices(angles, axis):
    """Convert rotation angles around `axis` to 3x3 rotation matrices.

    Args:
        angles: (N,) array of rotation angles (radian)
        axis: one of ('x', 'y', 'z')

    Returns:
        (N, 3, 3) array of rotation matrices.
    """
    return Rotation.from_euler(axis, angles).as_matrix()


def normalize(vs):
    """Normalize array along last axis."""
    norm = np.linalg.norm(vs, axis=-1)
    norm = np.where(norm > 0, norm, 1.0)
    return vs / norm[..., np.newaxis]


def isin(a, values):
    """Naive NumPy.isin analogue.

    For our usecases (>10^9 non-unique elements in `a`, <10^2 unique elements in tested `values`),
    NumPy.isin() takes same amount of time, but is 3x more memory-hungry.
    """
    a = np.asarray(a)
    result = np.full_like(a, False, dtype=bool)
    for v in set(values):
        result |= (a == v)
    return result


def euler2mat(az, ay, ax):
    """Build 3x3 rotation matrices from az, ay, ax rotation angles (in that order).

    Args:
        az: rotation angles around Z (Nx1 NumPy array; radians)
        ay: rotation angles around Y (Nx1 NumPy array; radians)
        ax: rotation angles around X (Nx1 NumPy array; radians)

    Returns:
        List with 3x3 rotation matrices corresponding to each of N angle triplets.

    See Also:
        https://en.wikipedia.org/wiki/Euler_angles#Rotation_matrix (R = X1 * Y2 * Z3)
    """
    c1, s1 = np.cos(ax), np.sin(ax)
    c2, s2 = np.cos(ay), np.sin(ay)
    c3, s3 = np.cos(az), np.sin(az)

    mm = np.array([
        [c2 * c3, -c2 * s3, s2],
        [c1 * s3 + c3 * s1 * s2, c1 * c3 - s1 * s2 * s3, -c2 * s1],
        [s1 * s3 - c1 * c3 * s2, c3 * s1 + c1 * s2 * s3, c1 * c2],
    ])

    return np.asarray([mm[..., i] for i in range(len(az))])


def mat2euler(mm):
    """Decompose 3x3 rotation matrices into az, ay, ax rotation angles (in that order).

    Args:
        List with 3x3 rotation matrices.

    Returns:
        az: rotation angles around Z (Nx1 NumPy array; radians)
        ay: rotation angles around Y (Nx1 NumPy array; radians)
        ax: rotation angles around X (Nx1 NumPy array; radians)

    See Also:
        https://en.wikipedia.org/wiki/Euler_angles#Rotation_matrix (R = X1 * Y2 * Z3)
    """
    assert len(mm.shape) == 3
    assert tuple(mm.shape[1:]) == (3, 3)

    az = np.full(mm.shape[0], np.nan)
    ay = np.full(mm.shape[0], np.nan)
    ax = np.full(mm.shape[0], np.nan)

    sin_ay = mm[:, 0, 2]

    mask1 = np.isclose(sin_ay, 1.0)
    az[mask1] = 0.0
    ay[mask1] = np.pi / 2
    ax[mask1] = np.arctan2(mm[:, 1, 0][mask1], mm[:, 1, 1][mask1])

    mask2 = np.isclose(sin_ay, -1.0)
    az[mask2] = 0.0
    ay[mask2] = -np.pi / 2
    ax[mask2] = np.arctan2(-mm[:, 1, 0][mask2], mm[:, 1, 1][mask2])

    mask3 = np.logical_not(np.logical_or(mask1, mask2))
    az[mask3] = np.arctan2(-mm[:, 0, 1][mask3], mm[:, 0, 0][mask3])
    ay[mask3] = np.arcsin(sin_ay[mask3])
    ax[mask3] = np.arctan2(-mm[:, 1, 2][mask3], mm[:, 2, 2][mask3])

    assert not np.any(np.isnan(az))
    assert not np.any(np.isnan(ay))
    assert not np.any(np.isnan(ax))

    return (az, ay, ax)


def voxel_intersection(line_segment, data, return_sub_segments=False):
    """Find voxels intersected by a given segment and cut the segment according to these voxels.

    Args:
        line_segment: The segment with the following form: [[x0, y0, z0], [x1, y1, z1]].
        data: The VoxelData object.
        return_sub_segments: If true the sub segments are also returned with the voxel indices.

    Returns:
        List of 3D indices.
        If `return_sub_segments` is `True`, the coordinates of the sub-segment points are returned.
    """
    # pylint: disable=too-many-locals

    # If the segment is outside the bounding box, then it does not intersect any voxels.
    if (line_segment < data.bbox).all() or (line_segment >= data.bbox).all():
        seg_point_indices = np.zeros((0, 3), dtype=np.result_type(0))
        if return_sub_segments:
            return seg_point_indices, np.reshape(line_segment, (1, -1))
        return seg_point_indices

    # The segment is clipped inside the global bbox.
    clipped_line_segment = np.clip(
        line_segment,
        a_min=data.bbox[0],
        a_max=np.nextafter(data.bbox[1], data.bbox[1] - 1),
    )

    start_pt, end_pt = clipped_line_segment

    def get_intersections(dst1, dst2):
        """Compute intersection point."""
        same_sign = np.sign(dst1) == np.sign(dst2)
        coplanar = (dst1 == 0) & (dst2 == 0)
        denomimator = dst2 - dst1
        denomimator = np.where(denomimator == 0, np.nan, denomimator)
        f = np.where(same_sign | coplanar, np.nan, -dst1 / denomimator)

        # Return the hit position.
        return start_pt + (end_pt - start_pt) * f[:, np.newaxis]

    # Compute the actual bbox of the segment
    bbox = np.sort(data.positions_to_indices(clipped_line_segment), axis=0)

    # Get the coordinates of the planes between voxels
    planes = [data.offset[i] + np.arange(bbox[0, i], bbox[1, i] + 1) * data.voxel_dimensions[i]
              for i in range(3)]

    # Get the coordinates of the intersection points
    seg_points = np.vstack(
        [get_intersections(start_pt[i] - planes[i], end_pt[i] - planes[i])
         for i in range(3)]
    )

    # Build the sub-segment coordinate array
    seg_points = seg_points[~np.isnan(seg_points).all(axis=1)]
    seg_points = np.unique(seg_points, axis=0)

    # Check how the points are ordered along each axis
    xyz_ascending = np.sign(end_pt - start_pt)

    # Remove duplicated points when the extremities of the segment are on a voxel boundary, except
    # if they are in the ascending quadrant.
    new_pts = [seg_points]
    if xyz_ascending.sum() >= 0:
        new_pts = new_pts + [end_pt]
        seg_pt_start = (seg_points == start_pt).all(axis=1)
        if (
            not seg_pt_start.any()
            and not any(len(planes[i]) > 1 and (start_pt[i] == planes[i]).any() for i in range(3))
        ):
            new_pts = [start_pt] + new_pts
    else:
        new_pts = [start_pt] + new_pts
        seg_pt_end = (seg_points == end_pt).all(axis=1)
        if (
            not seg_pt_end.any()
            and not any(len(planes[i]) > 1 and (end_pt[i] == planes[i]).any() for i in range(3))
        ):
            new_pts = new_pts + [end_pt]

    # Build the sub-segment points
    seg_points = np.vstack(new_pts)

    # Sort the sub-segment points so their order is consistent with the given segment, i.e. the
    # first point is the start point, the second pt is the closest to the start pt, etc.
    ind = np.lexsort((seg_points[:, 0], seg_points[:, 1], seg_points[:, 2]))

    if not (
        xyz_ascending[0] == 1
        or (xyz_ascending[0] == 0 and xyz_ascending[1] == 1)
        or (xyz_ascending[0] == 0 and xyz_ascending[1] == 0 and xyz_ascending[2] == 1)
    ):
        ind = ind[::-1]

    seg_points = seg_points[ind, :]

    # Find the intersection indices
    seg_point_indices = data.positions_to_indices((seg_points[:-1, :] + seg_points[1:, :]) / 2.)

    if return_sub_segments:
        sub_segments = np.hstack([seg_points[:-1], seg_points[1:]])
        return seg_point_indices, sub_segments

    return seg_point_indices



#!/usr/bin/env python
"""Distribution configuration."""

from setuptools import setup, find_packages

with open("README.rst", encoding='utf-8') as f:
    README = f.read()

setup(
    name='voxcell',
    author="Blue Brain Project, EPFL",
    description='Voxcell is a small library to handle probability'
                ' distributions that have a spatial component and to use them'
                ' to build collection of cells in space.',
    long_description=README,
    long_description_content_type="text/x-rst",
    url="https://github.com/BlueBrain/voxcell",
    download_url="https://github.com/BlueBrain/voxcell",
    license='Apache-2',
    install_requires=[
        'h5py>=3.1.0',
        'numpy>=1.9',
        'pandas>=0.24.2',
        'pynrrd>=1.1.0',
        'requests>=2.18',
        'scipy>=1.2.0',
    ],
    packages=find_packages(),
    python_requires='>=3.8',

    setup_requires=[
        'setuptools_scm',
    ],
    use_scm_version={
        "local_scheme": "no-local-version",
    },

    extras_require={
        'docs': [
            'sphinx-bluebrain-theme',
        ],
    },

    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
    ],
)



import numpy as np
import pandas as pd
from numpy.testing import assert_array_equal
from pandas.testing import assert_frame_equal

from voxcell.traits import SpatialDistribution, _drop_duplicate_columns
from voxcell.voxel_data import VoxelData

NOFIELD = np.array([])


def test_split_distribution_collection_empty_0():
    probabilities = pd.DataFrame()
    attributes = ['name']
    traits = pd.DataFrame(columns=attributes)

    assert (SpatialDistribution(NOFIELD, probabilities, traits).split(attributes) == {})


def test_split_distribution_collection_empty_1():
    traits = pd.DataFrame([{'name': 'a'}])
    distributions = pd.DataFrame(index=traits.index)
    attributes = 'name'

    res = SpatialDistribution(NOFIELD, distributions, traits).split(attributes)
    assert sorted(res.keys()) == ['a']
    assert_frame_equal(res['a'].distributions, distributions)
    assert_frame_equal(res['a'].traits, traits)


def test_split_distribution_collection_empty_2():
    traits = pd.DataFrame([{'name': 'a'}, {'name': 'b'}])
    distributions = pd.DataFrame(index=traits.index)
    attributes = 'name'

    res = SpatialDistribution(NOFIELD, distributions, traits).split(attributes)
    assert sorted(res.keys()) == ['a', 'b']
    assert_frame_equal(res['a'].distributions, pd.DataFrame(index=[0]))
    assert_frame_equal(res['b'].distributions, pd.DataFrame(index=[1]))
    # {('a',): SD(NOFIELD, [], traits), ('b',): SD(NOFIELD, [], traits)}


def test_split_distribution_collection_single_0():
    distributions = pd.DataFrame([{0: 1.}])
    traits = pd.DataFrame([{'name': 'a'}])
    attributes = 'name'

    # {('a',): SD(NOFIELD, [{0: 1}], traits)})

    res = SpatialDistribution(NOFIELD, distributions, traits).split(attributes)
    assert sorted(res.keys()) == ['a']
    assert_frame_equal(res['a'].distributions, pd.DataFrame([1.]))
    assert_frame_equal(res['a'].traits, traits)


def test_split_distribution_collection_single_1():
    distributions = pd.DataFrame([0.25, 0.75])
    traits = pd.DataFrame([{'name': 'a'}, {'name': 'b'}])
    attributes = 'name'

    # {('a',): SD(NOFIELD, [{0: 1.0}], traits), ('b',): SD(NOFIELD, [{1: 1.0}], traits)})
    res = SpatialDistribution(NOFIELD, distributions, traits).split(attributes)
    assert sorted(res.keys()) == ['a', 'b']
    assert_frame_equal(res['a'].distributions, pd.DataFrame([1.]))
    assert_frame_equal(res['b'].distributions, pd.DataFrame([1.], index=[1]))


def test_split_distribution_collection_single_2():
    distributions = pd.DataFrame([0.2, 0.4, 0.4])
    traits = pd.DataFrame([{'name': 'a'}, {'name': 'b'}, {'name': 'b'}])
    attributes = 'name'

    res = SpatialDistribution(NOFIELD, distributions, traits).split(attributes)
    assert sorted(res.keys()) == ['a', 'b']
    assert_frame_equal(res['a'].distributions, pd.DataFrame([1.]))
    assert_frame_equal(res['b'].distributions, pd.DataFrame([0.5, 0.5], index=[1, 2]))


def test_split_distribution_collection_multiattr_0():
    distributions = pd.DataFrame([0.2, 0.4, 0.4])
    traits = pd.DataFrame([{'name': 'a', 'type': 'x'},
                           {'name': 'b', 'type': 'y'},
                           {'name': 'b', 'type': 'y'}])

    attributes = ['name', 'type']

    # {('a', 'x'): SD(NOFIELD, [{0: 1.0}], traits),
    #  ('b', 'y'): SD(NOFIELD, [{1: 0.5, 2: 0.5}], traits)})

    res = SpatialDistribution(NOFIELD, distributions, traits).split(attributes)
    assert sorted(res.keys()) == [('a', 'x'), ('b', 'y')]
    assert_frame_equal(res[('a', 'x')].distributions, pd.DataFrame([1.0]))
    assert_frame_equal(res[('b', 'y')].distributions, pd.DataFrame([0.5, 0.5], index=[1, 2]))


def test_split_distribution_collection_multiattr_1():
    distributions = pd.DataFrame([0.2, 0.4, 0.4])
    traits = pd.DataFrame([{'name': 'a', 'type': 'x'},
                           {'name': 'b', 'type': 'x'},
                           {'name': 'b', 'type': 'y'}])

    attributes = ['name', 'type']

    res = SpatialDistribution(NOFIELD, distributions, traits).split(attributes)
    assert sorted(res.keys()) == [('a', 'x'), ('b', 'x'), ('b', 'y')]
    assert_frame_equal(res[('a', 'x')].distributions, pd.DataFrame([1.]))
    assert_frame_equal(res[('b', 'x')].distributions, pd.DataFrame([1.], index=[1]))
    assert_frame_equal(res[('b', 'y')].distributions, pd.DataFrame([1.], index=[2]))

    # {('a', 'x'): SD(NOFIELD, [{0: 1.0}], traits),
    #  ('b', 'x'): SD(NOFIELD, [{1: 1.0}], traits),
    #  ('b', 'y'): SD(NOFIELD, [{2: 1.0}], traits)})


def test_reduce_distribution_collection_empty_0():
    sd = SpatialDistribution(NOFIELD, pd.DataFrame(), pd.DataFrame(columns=['name']))
    r = sd.reduce('name')
    assert_frame_equal(r.distributions, pd.DataFrame())


def test_reduce_distribution_collection_empty_1():
    traits = pd.DataFrame([{'name': 'a', 'type': 'x'}])
    dists = pd.DataFrame(index=traits.index)
    r = SpatialDistribution(NOFIELD, dists, traits).reduce('type')
    assert_frame_equal(r.distributions, dists)


def test_reduce_distribution_collection_0():
    traits = pd.DataFrame([{'name': 'a', 'type': 'x'},
                           {'name': 'b', 'type': 'x'}])

    dists = pd.DataFrame([0.25, 0.75])

    r = SpatialDistribution(NOFIELD, dists, traits).reduce('type')
    assert_frame_equal(r.distributions,
                       pd.DataFrame([1.]))


def test_reduce_distribution_collection_1():
    traits = pd.DataFrame([{'name': 'a', 'type': 'x'},
                          {'name': 'c', 'type': 'y'}])

    dists = pd.DataFrame([0.75, 0.25])

    r = SpatialDistribution(NOFIELD, dists, traits).reduce('type')
    assert_frame_equal(r.traits,
                       pd.DataFrame({'type': ['x', 'y']}))
    assert_frame_equal(r.distributions,
                       pd.DataFrame([0.75, 0.25]))


def test_reduce_distribution_collection_2():
    traits = pd.DataFrame([{'name': 'a', 'type': 'x', 'color': 0},
                           {'name': 'b', 'type': 'x', 'color': 1},
                           {'name': 'c', 'type': 'y', 'color': 2},
                           {'name': 'd', 'type': 'y', 'color': 3}])

    dists = pd.DataFrame([0.1, 0.2, 0.3, 0.4])

    r = SpatialDistribution(NOFIELD, dists, traits).reduce('type')
    assert_frame_equal(r.traits,
                       pd.DataFrame({'type': ['x', 'y']}))
    assert_frame_equal(r.distributions,
                       pd.DataFrame([0.3, 0.7]))

    r = SpatialDistribution(NOFIELD, dists, traits).reduce('name')
    assert_frame_equal(r.traits,
                       pd.DataFrame({'name': ['a', 'b', 'c', 'd']}))
    assert_frame_equal(r.distributions,
                       pd.DataFrame([0.1, 0.2, 0.3, 0.4]))


def check_assign_conditional(preassigned, expected):
    sdist = SpatialDistribution(field=VoxelData(np.array([[[0]]]), (25, 25, 25)),
                                distributions=pd.DataFrame([0.75, 0.25, 0.0]),
                                traits=pd.DataFrame([{'name': 'a', 'type': 'x', 'color': 0},
                                                     {'name': 'b', 'type': 'y', 'color': 0},
                                                     {'name': 'c', 'type': 'z', 'color': 1}]))

    chosen = sdist.assign_conditional(np.array([[1, 1, 1]]), preassigned)

    assert chosen == np.array([expected])


def test_assign_conditional_single_series():
    check_assign_conditional(pd.DataFrame({'type': ['x']}).type, 0)
    check_assign_conditional(pd.DataFrame({'type': ['y']}).type, 1)


def test_assign_conditional_single_dataframe():
    check_assign_conditional(pd.DataFrame({'type': ['x']}), 0)
    check_assign_conditional(pd.DataFrame({'type': ['y']}), 1)


def test_assign_conditional_multiple():
    check_assign_conditional(pd.DataFrame({'type': ['x'], 'color': [0]}), 0)
    check_assign_conditional(pd.DataFrame({'type': ['y'], 'color': [0]}), 1)


def test_assign_conditional_single_unknown():
    preassigned = pd.DataFrame({'type': ['unknown']})
    check_assign_conditional(preassigned, -1)


def test_assign_conditional_single_impossible():
    preassigned = pd.DataFrame({'type': ['z']})
    check_assign_conditional(preassigned, -1)


def test_assign_conditional_multiple_unknown():
    preassigned = pd.DataFrame({'type': ['unknown'], 'color': [0]})
    check_assign_conditional(preassigned, -1)


def test_assign_conditional_multiple_impossible_0():
    # probability of this combination is zero
    preassigned = pd.DataFrame({'type': ['z'], 'color': [1]})
    check_assign_conditional(preassigned, -1)


def test_assign_conditional_multiple_impossible_1():
    # preassigned values exist but combination does not
    preassigned = pd.DataFrame({'name': ['a'], 'type': ['y']})
    check_assign_conditional(preassigned, -1)


def test_assign_atlas_int8():
    prob = np.zeros(1000)
    prob[999] = 1.0
    sdist = SpatialDistribution(
        field=VoxelData(np.array([1], dtype=np.int8), (25,)),
        distributions=pd.DataFrame({1: prob}),
        traits=None
    )
    assert sdist.assign([[0]]) == 999

def test_collect():
    sdist = SpatialDistribution(
        field=VoxelData(np.array([[[0]]]), (25, 25, 25)),
        distributions=pd.DataFrame([0.75, 0.25, 0.0]),
        traits=pd.DataFrame([
            {'name': 'a', 'type': 'x', 'color': 0},
            {'name': 'b', 'type': 'y', 'color': 0},
            {'name': 'c', 'type': 'z', 'color': 1}
        ])
    )

    positions = np.array([[1, 1, 1]])
    preassigned = pd.DataFrame([{
        'type': 'y'
    }])

    expected = pd.DataFrame([{
        'color': 0
    }])

    result = sdist.collect(positions, preassigned, names=['color'])
    assert_frame_equal(result, expected)


def test_drop_duplicate_columns_0():
    df = pd.DataFrame([[0.4, 0.2],
                       [0.6, 0.8]])

    unique, inverse = _drop_duplicate_columns(df)

    expected = pd.DataFrame([[0.4, 0.2],
                             [0.6, 0.8]])
    assert_frame_equal(unique, expected)

    assert_array_equal(inverse, np.array([0, 1]))


def test_drop_duplicate_columns_single_0():
    df = pd.DataFrame([[0.4, 0.4], [0.6, 0.6]])

    unique, inverse = _drop_duplicate_columns(df)

    assert_frame_equal(unique, pd.DataFrame([[0.4], [0.6]]))
    assert_array_equal(inverse, np.array([0, 0]))


def test_drop_duplicate_columns_single_rounded_0():
    df = pd.DataFrame([[0.39999999999999997, 0.40000000000000002], [0.6, 0.6]])

    unique, inverse = _drop_duplicate_columns(df, decimals=100)

    # no change
    assert_frame_equal(unique, pd.DataFrame([[0.39999999999999997, 0.40000000000000002],
                                             [0.6, 0.6]]))
    assert_array_equal(inverse, np.array([0, 1]))


def test_drop_duplicate_columns_single_rounded_1():
    df = pd.DataFrame([[0.39999999999999997, 0.40000000000000002], [0.6, 0.6]])

    unique, inverse = _drop_duplicate_columns(df, decimals=20)

    # dropped the second column
    assert_frame_equal(unique, pd.DataFrame([[0.4],
                                             [0.6]]))
    assert_array_equal(inverse, np.array([0, 0]))


def test_drop_duplicate_columns_1():
    df = pd.DataFrame([[0.4, 0.2, 0.2, 0.4, 0.2],
                       [0.6, 0.8, 0.8, 0.6, 0.8]])

    unique, inverse = _drop_duplicate_columns(df)

    expected = pd.DataFrame([[0.4, 0.2],
                             [0.6, 0.8]])
    assert_frame_equal(unique, expected)

    assert_array_equal(inverse, np.array([0, 1, 1, 0, 1]))


def test_drop_duplicates():

    sd = SpatialDistribution(
        field=VoxelData(np.array([0, 0, 1, 1, 2, 2, 3, 3, 4, 4]), (1,)),
        distributions=pd.DataFrame([[0.4, 0.2, 0.2, 0.4, 0.2],
                                    [0.6, 0.8, 0.8, 0.6, 0.8]]),
        traits=pd.DataFrame({'size': ['Large', 'Small'], 'speed': ['Fast', 'Slow']}))

    result = sd.drop_duplicates()

    assert_array_equal(result.field.raw, [0, 0, 1, 1, 1, 1, 0, 0, 1, 1])
    assert_frame_equal(result.distributions, pd.DataFrame([[0.4, 0.2], [0.6, 0.8]]))

    # original is not modified
    assert_array_equal(sd.field.raw, [0, 0, 1, 1, 2, 2, 3, 3, 4, 4])


def test_get_probability_field():

    sd = SpatialDistribution(
        field=VoxelData(np.array([0, 0, -1, 1, 1]), (1,)),
        distributions=pd.DataFrame([[0.4, 0.2],
                                    [0.6, 0.8]]),
        traits=pd.DataFrame({'size': ['Large', 'Small']}))

    assert_array_equal(sd.get_probability_field('size', 'Large').raw, [0.4, 0.4, -1, 0.2, 0.2])
    assert_array_equal(sd.get_probability_field('size', 'Small').raw, [0.6, 0.6, -1, 0.8, 0.8])


def test_from_probability_field():

    sd = SpatialDistribution.from_probability_field(
        VoxelData(np.array([0.4, 0.4, -1, 0.2, 0.2]), (1,)),
        'size', 'Large', 'Small'
    )

    # same as previous test but the columns in distributions are swapped
    # the reason for this is internal to numpy.unique but it doesn't really affect the result
    assert_array_equal(sd.field.raw, [1, 1, -1, 0, 0])
    assert_frame_equal(sd.distributions, pd.DataFrame([[0.2, 0.4], [0.8, 0.6]]))
    assert_frame_equal(sd.traits, pd.DataFrame({'size': ['Large', 'Small']}))

    # round trip
    assert_array_equal(sd.get_probability_field('size', 'Large').raw, [0.4, 0.4, -1, 0.2, 0.2])
    assert_array_equal(sd.get_probability_field('size', 'Small').raw, [0.6, 0.6, -1, 0.8, 0.8])
